__all__ = ["kill_me"]

import os
from typing import List

import psutil

from neptune_scale.util.envs import SUBPROCESS_KILL_TIMEOUT

KILL_TIMEOUT = int(os.getenv(SUBPROCESS_KILL_TIMEOUT, "5"))


def kill_me() -> None:
    process = psutil.Process(os.getpid())
    try:
        children = _get_process_children(process)
    except psutil.NoSuchProcess:
        children = []

    for child_proc in children:
        _terminate(child_proc)
    _, alive = psutil.wait_procs(children, timeout=KILL_TIMEOUT)
    for child_proc in alive:
        _kill(child_proc)
    # finish with terminating self
    _terminate(process)


def _terminate(process: psutil.Process) -> None:
    try:
        process.terminate()
    except psutil.NoSuchProcess:
        pass


def _kill(process: psutil.Process) -> None:
    try:
        if process.is_running():
            process.kill()
    except psutil.NoSuchProcess:
        pass


def _get_process_children(process: psutil.Process) -> List[psutil.Process]:
    try:
        return process.children(recursive=True)
    except psutil.NoSuchProcess:
        return []



from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)
from types import TracebackType
from typing import (
    Optional,
    Type,
)


class AutoCloseable(ABC):
    def __enter__(self) -> AutoCloseable:
        return self

    @abstractmethod
    def close(self) -> None: ...

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()


class Resource(AutoCloseable):
    def cleanup(self) -> None:
        pass

    def flush(self) -> None:
        pass

    def close(self) -> None:
        self.flush()


class WithResources(Resource):
    @property
    @abstractmethod
    def resources(self) -> tuple[Resource, ...]: ...

    def flush(self) -> None:
        for resource in self.resources:
            resource.flush()

    def close(self) -> None:
        for resource in self.resources:
            resource.close()

    def cleanup(self) -> None:
        for resource in self.resources:
            resource.cleanup()



import multiprocessing
from types import TracebackType
from typing import (
    Callable,
    Generic,
    Optional,
    Type,
    TypeVar,
    cast,
)

__all__ = ("SharedInt", "SharedFloat", "SharedVar")

T = TypeVar("T", int, float)


class SharedVar(Generic[T]):
    """A convenience wrapper around multiprocessing.Value and multiprocessing.Condition.

    Access the Value using the `value` property. Other methods, as well as context management
    delegate to the Condition object.

    Typical usage:
        var = SharedInt(0)

        # In one process
        with var:
            var.value += 1
            var.notify() # Notify the waiting process

        # In another process
        with var:
            var.wait() # Wait until the value is changed
            print(var.value)
    """

    def __init__(self, typecode_or_type: str, initial_value: T):
        self._value = multiprocessing.Value(typecode_or_type, initial_value)
        self._condition = multiprocessing.Condition(self._value.get_lock())

    @property
    def value(self) -> T:
        with self._condition:
            return cast(T, self._value.value)

    @value.setter
    def value(self, new_value: T) -> None:
        with self._condition:
            self._value.value = new_value

    def wait(self, timeout: Optional[float] = None) -> None:
        with self._condition:
            self._condition.wait(timeout)

    def wait_for(self, predicate: Callable[[], bool], timeout: Optional[float] = None) -> None:
        with self._condition:
            self._condition.wait_for(predicate, timeout)

    def notify_all(self) -> None:
        with self._condition:
            self._condition.notify_all()

    def __enter__(self) -> "SharedVar[T]":
        self._condition.__enter__()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._condition.__exit__(exc_type, exc_value, traceback)


class SharedInt(SharedVar[int]):
    def __init__(self, initial_value: int) -> None:
        super().__init__("i", initial_value)


class SharedFloat(SharedVar[float]):
    def __init__(self, initial_value: float) -> None:
        super().__init__("d", initial_value)



from __future__ import annotations

__all__ = ("ErrorsQueue", "ErrorsMonitor")

import multiprocessing
import queue
import time
from typing import (
    Callable,
    Optional,
    Type,
)

from neptune_scale.exceptions import (
    NeptuneAsyncLagThresholdExceeded,
    NeptuneConnectionLostError,
    NeptuneOperationsQueueMaxSizeExceeded,
    NeptuneScaleError,
    NeptuneScaleWarning,
    NeptuneUnexpectedError,
)
from neptune_scale.sync.parameters import ERRORS_MONITOR_THREAD_SLEEP_TIME
from neptune_scale.util import get_logger
from neptune_scale.util.abstract import Resource
from neptune_scale.util.daemon import Daemon
from neptune_scale.util.process_killer import kill_me

logger = get_logger()


class ErrorsQueue(Resource):
    def __init__(self) -> None:
        self._errors_queue: multiprocessing.Queue[BaseException] = multiprocessing.Queue()

    def put(self, error: BaseException) -> None:
        self._errors_queue.put(error)

    def get(self, block: bool = True, timeout: Optional[float] = None) -> BaseException:
        return self._errors_queue.get(block=block, timeout=timeout)

    def close(self) -> None:
        self._errors_queue.close()
        # This is needed to avoid hanging the main process
        self._errors_queue.cancel_join_thread()


def default_error_callback(error: BaseException, last_seen_at: Optional[float]) -> None:
    logger.error(f"Terminating the process due to an error: {error}")
    kill_me()


def default_network_error_callback(error: BaseException, last_seen_at: Optional[float]) -> None:
    if last_seen_at is None or time.time() - last_seen_at > 5:
        logger.warning(f"A network error occurred: {error}. Retrying...")


def default_max_queue_size_exceeded_callback(error: BaseException, last_raised_at: Optional[float]) -> None:
    if last_raised_at is None or time.time() - last_raised_at > 5:
        logger.warning(f"Pending operations queue size exceeded: {error}. Retrying...")


def default_warning_callback(error: BaseException, last_seen_at: Optional[float]) -> None:
    logger.warning(error)


class ErrorsMonitor(Daemon, Resource):
    def __init__(
        self,
        errors_queue: ErrorsQueue,
        on_queue_full_callback: Optional[Callable[[BaseException, Optional[float]], None]] = None,
        on_async_lag_callback: Optional[Callable[[], None]] = None,
        on_network_error_callback: Optional[Callable[[BaseException, Optional[float]], None]] = None,
        on_error_callback: Optional[Callable[[BaseException, Optional[float]], None]] = None,
        on_warning_callback: Optional[Callable[[BaseException, Optional[float]], None]] = None,
    ):
        super().__init__(name="ErrorsMonitor", sleep_time=ERRORS_MONITOR_THREAD_SLEEP_TIME)

        self._errors_queue: ErrorsQueue = errors_queue
        self._on_queue_full_callback: Callable[[BaseException, Optional[float]], None] = (
            on_queue_full_callback or default_max_queue_size_exceeded_callback
        )
        self._on_async_lag_callback: Callable[[], None] = on_async_lag_callback or (lambda: None)
        self._on_network_error_callback: Callable[[BaseException, Optional[float]], None] = (
            on_network_error_callback or default_network_error_callback
        )
        self._on_error_callback: Callable[[BaseException, Optional[float]], None] = (
            on_error_callback or default_error_callback
        )
        self._on_warning_callback: Callable[[BaseException, Optional[float]], None] = (
            on_warning_callback or default_warning_callback
        )

        self._last_raised_timestamps: dict[Type[BaseException], float] = {}

    def get_next(self) -> Optional[BaseException]:
        try:
            return self._errors_queue.get(block=False)
        except queue.Empty:
            return None

    def work(self) -> None:
        while (error := self.get_next()) is not None:
            last_raised_at = self._last_raised_timestamps.get(type(error), None)
            self._last_raised_timestamps[type(error)] = time.time()

            try:
                if isinstance(error, NeptuneOperationsQueueMaxSizeExceeded):
                    self._on_queue_full_callback(error, last_raised_at)
                elif isinstance(error, NeptuneConnectionLostError):
                    self._on_network_error_callback(error, last_raised_at)
                elif isinstance(error, NeptuneScaleWarning):
                    self._on_warning_callback(error, last_raised_at)
                elif isinstance(error, NeptuneScaleError):
                    self._on_error_callback(error, last_raised_at)
                elif isinstance(error, NeptuneAsyncLagThresholdExceeded):
                    self._on_async_lag_callback()
                else:
                    self._on_error_callback(NeptuneUnexpectedError(reason=str(error)), last_raised_at)
            except Exception as e:
                # Don't let user errors kill the process
                logger.error(f"An exception occurred in user callback function: {e}")



__all__ = ("BatchedOperations", "SingleOperation")

from typing import (
    NamedTuple,
    Optional,
)


class BatchedOperations(NamedTuple):
    # Operation identifier of the last operation in the batch
    sequence_id: int
    # Timestamp of the last operation in the batch
    timestamp: float
    # Protobuf serialized (RunOperationBatch)
    operation: bytes


class SingleOperation(NamedTuple):
    # Operation identifier
    sequence_id: int
    # Timestamp of the operation being enqueued
    timestamp: float
    # Protobuf serialized (RunOperation)
    operation: bytes
    # Whether the operation is batchable. Eg. metadata updates are, while run creations are not.
    is_batchable: bool
    # Size of the metadata in the operation (without project, family, run_id etc.)
    metadata_size: Optional[int]
    # Update metadata key
    batch_key: Optional[float]



__all__ = [
    "Daemon",
    "get_logger",
    "ProcessLink",
    "SharedFloat",
    "SharedInt",
    "SharedVar",
]

from neptune_scale.util.daemon import Daemon
from neptune_scale.util.logger import get_logger
from neptune_scale.util.process_link import ProcessLink
from neptune_scale.util.shared_var import (
    SharedFloat,
    SharedInt,
    SharedVar,
)



__all__ = ("get_logger",)

import logging
import os

from neptune_scale.util.envs import DEBUG_MODE
from neptune_scale.util.styles import (
    STYLES,
    ensure_style_detected,
)


class NeptuneWarning(Warning): ...


LOG_FORMAT = "{blue}%(name)s{end}:{bold}%(levelname)s{end}: %(message)s"
DEBUG_FORMAT = (
    "%(asctime)s:%(name)s:%(levelname)s:%(processName)s(%(process)d):%(threadName)s:%(filename)s:"
    "%(funcName)s():%(lineno)d: %(message)s"
)


def get_logger() -> logging.Logger:
    """Use in modules to get the root Neptune logger"""

    logger = logging.getLogger("neptune")

    # If the user has also imported `neptune-fetcher` the root logger will already be initialized.
    # We want our handlers to take precedence. We will remove all handlers and add our own.
    if logger.hasHandlers():
        # Already initialized by us
        if hasattr(logger, "__neptune_scale"):
            return logger

        # Clear handlers and proceed with initialization
        logger.handlers.clear()

    logger.setLevel(logging.INFO)
    ensure_style_detected()

    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(logging.Formatter(LOG_FORMAT.format(**STYLES)))
    logger.addHandler(stream_handler)

    if os.environ.get(DEBUG_MODE, "False").lower() in ("true", "1"):
        logger.setLevel(logging.DEBUG)

        file_handler = logging.FileHandler(f"neptune.{os.getpid()}.log")
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter(DEBUG_FORMAT))
        logger.addHandler(file_handler)

    logger.__neptune_scale = True  # type: ignore

    return logger



import os

PROJECT_ENV_NAME = "NEPTUNE_PROJECT"

API_TOKEN_ENV_NAME = "NEPTUNE_API_TOKEN"

DISABLE_COLORS = "NEPTUNE_DISABLE_COLORS"

DEBUG_MODE = "NEPTUNE_DEBUG_MODE"

SUBPROCESS_KILL_TIMEOUT = "NEPTUNE_SUBPROCESS_KILL_TIMEOUT"

ALLOW_SELF_SIGNED_CERTIFICATE = "NEPTUNE_ALLOW_SELF_SIGNED_CERTIFICATE"
SKIP_NON_FINITE_METRICS = "NEPTUNE_SKIP_NON_FINITE_METRICS"


def get_bool(name: str, default: bool) -> bool:
    return os.getenv(name, str(default)).lower() in ("true", "1")
